import abc
from typing import Any, Dict, Optional

from tgedr.dataops.chain import Chain


class SourceException(Exception):
    pass


class SourceInterface(metaclass=abc.ABCMeta):
    """
    def get(self, context: Optional[Dict[str, Any]] = None) -> Any:
        raise NotImplementedError()
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, "get") and callable(subclass.get) or NotImplemented


@SourceInterface.register
class Source(abc.ABC):
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        self._config = config

    @abc.abstractmethod
    def get(self, context: Optional[Dict[str, Any]] = None) -> Any:
        raise NotImplementedError()


@SourceInterface.register
class SourceChain(Chain, abc.ABC):
    def execute(self, context: Optional[Dict[str, Any]] = None) -> Any:
        return self.get(context=context)



import abc
from typing import Any, Dict, Optional


class ProcessorException(Exception):
    pass


class ProcessorInterface(metaclass=abc.ABCMeta):
    """
    def process(self, context: Optional[Dict[str, Any]] = None) -> Any:
        raise NotImplementedError()
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, "process") and callable(subclass.process) or NotImplemented


@ProcessorInterface.register
class Processor(abc.ABC):
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        self._config = config

    @abc.abstractmethod
    def process(self, context: Optional[Dict[str, Any]] = None) -> Any:
        raise NotImplementedError()



from abc import ABC, abstractmethod
import logging
import os
from typing import Any, Dict, List, Optional
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from tgedr.dataops.store.store import Store, StoreException


logger = logging.getLogger(__name__)


class FsSinglePartitionParquetStore(Store, ABC):
    @property
    @abstractmethod
    def fs(self):
        raise NotImplementedError()

    @abstractmethod
    def _rmdir(self, key):
        raise NotImplementedError()

    @abstractmethod
    def _exists(self, key) -> bool:
        raise NotImplementedError()

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        Store.__init__(self, config)
        self._fs = None

    def get(self, key: str, use_legacy_dataset: bool = False) -> pd.DataFrame:
        logger.info(f"[get|in] ({key}, {use_legacy_dataset})")
        result = pq.read_table(key, use_legacy_dataset=use_legacy_dataset, filesystem=self.fs).to_pandas()
        logger.info(f"[get|out] => {result}")
        return result

    def delete(
        self,
        key: str,
        partition_field: Optional[str] = None,
        partitions: Optional[List[str]] = None,
        kv_dict: Optional[Dict[str, List[Any]]] = None,
    ):
        logger.info(f"[delete|in] ({key}, {partition_field}, {partitions}, {kv_dict})")

        if partitions is not None and partition_field is not None:
            self._remove_partitions(key, partition_field=partition_field, partition_values=partitions)
        elif kv_dict is not None and partition_field is not None:
            df0 = self.get(key)
            df = pd.DataFrame(kv_dict)
            joined_index = pd.merge(df0, df, on=list(kv_dict.keys()), how="inner").index
            df0.drop(joined_index, inplace=True)
            self.save(df0, key, partition_field=partition_field)
        else:
            self._rmdir(key)

        logger.info("[delete|out]")

    def save(
        self,
        df: pd.DataFrame,
        key: str,
        partition_field: Optional[str] = None,
        append: bool = False,
        replace_partitions: bool = False,
    ):
        logger.info(f"[save|in] ({df}, {key}, {partition_field}, {append}, {replace_partitions})")
        if replace_partitions and append:
            raise StoreException(f"cannot request for replace_partitions and append at the same time")
        if append:
            pq.write_to_dataset(
                pa.Table.from_pandas(df), root_path=key, partition_cols=[partition_field], filesystem=self.fs
            )
        elif replace_partitions:
            partitions = df[partition_field].unique().tolist()
            self._remove_partitions(key, partition_field, partitions)
            pq.write_to_dataset(
                pa.Table.from_pandas(df),
                root_path=key,
                partition_cols=[partition_field],
                existing_data_behavior="delete_matching",
                filesystem=self.fs,
            )
        else:
            self.delete(key)
            pq.write_to_dataset(
                pa.Table.from_pandas(df),
                root_path=key,
                partition_cols=[partition_field],
                existing_data_behavior="delete_matching",
                filesystem=self.fs,
            )
        logger.info("[save|out]")

    def _remove_partitions(self, key: str, partition_field: str, partition_values: List[str]):
        logger.debug(f"[_remove_partitions|in] ({key}, {partition_field}, {partition_values})")

        for partition_value in partition_values:
            partition_key = f"{partition_field}={partition_value}"
            partition_path = os.path.join(key, partition_key)
            self._rmdir(partition_path)

        logger.debug("[_remove_partitions|out]")

    def update(
        self,
        df: pd.DataFrame,
        key: str,
        key_fields: List[str],
        partition_field: Optional[str] = None,
    ):
        logger.info(f"[save|in] ({df}, {key}, {key_fields}, {partition_field})")

        df0 = self.get(key)
        joined_index = pd.merge(df0, df, on=key_fields, how="inner").index
        df0.loc[joined_index] = df
        self.save(df0, key, partition_field=partition_field)

        logger.info(f"[save|out]")



import logging
import os
from typing import Any, Dict, List, Optional
from tgedr.dataops.commons.s3_connector import S3Connector
from tgedr.dataops.sink.sink import Sink, SinkException


logger = logging.getLogger(__name__)


class S3FileSink(Sink, S3Connector):
    CONTEXT_SOURCE_PATH = "source"
    CONTEXT_TARGET_PATH = "target"

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        Sink.__init__(self, config=config)
        S3Connector.__init__(self)

    def put(self, context: Optional[Dict[str, Any]] = None) -> Any:
        logger.info(f"[put|in] ({context})")

        if self.CONTEXT_SOURCE_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_SOURCE_PATH}")
        if self.CONTEXT_TARGET_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_TARGET_PATH}")

        target = context[self.CONTEXT_TARGET_PATH]
        target_elements = target.split("/")
        bucket = target_elements[0]
        key = "/".join(target_elements[1:])

        source = context[self.CONTEXT_SOURCE_PATH]

        if os.path.isdir(source):
            files: List[str] = [f for f in os.listdir(source) if not os.path.isdir(os.path.join(source, f))]
            for file in files:
                target_key = os.path.join(key, file)
                source_file = os.path.join(source, file)
                logger.info(f"[put] uploading file {source_file} to key {target_key}")
                self._client.upload_file(Filename=source_file, Bucket=bucket, Key=target_key)
        else:
            logger.info(f"[put] uploading single file {source} to key {key}")
            self._client.upload_file(Filename=source, Bucket=bucket, Key=key)

        logger.info("[put|out]")



import abc
from typing import Any, Dict, Optional


class StoreException(Exception):
    pass


class StoreInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "get")
            and callable(subclass.get)
            and hasattr(subclass, "delete")
            and callable(subclass.delete)
            and hasattr(subclass, "save")
            and callable(subclass.save)
            and hasattr(subclass, "update")
            and callable(subclass.update)
        ) or NotImplemented


@StoreInterface.register
class Store(abc.ABC):
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        self._config = config

    @abc.abstractmethod
    def get(self, key: str, **kwargs) -> Any:
        raise NotImplementedError()

    @abc.abstractmethod
    def delete(self, key: str, **kwargs) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def save(self, df: Any, key: str, **kwargs):
        raise NotImplementedError()

    @abc.abstractmethod
    def update(self, df: Any, key: str, **kwargs):
        raise NotImplementedError()



import logging
import os
import shutil
from typing import Any, Dict, List, Optional
from tgedr.dataops.source.source import Source, SourceException


logger = logging.getLogger(__name__)


class LocalFsFileSource(Source):
    CONTEXT_KEY_SOURCE = "source"
    CONTEXT_KEY_TARGET = "target"
    CONTEXT_KEY_SUFFIX = "file_suffix"
    __DEFAULT_SUFFIX = ".txt"

    def get(self, context: Optional[Dict[str, Any]] = None) -> Any:
        logger.info(f"[get|in] ({context})")

        if self.CONTEXT_KEY_SOURCE not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_SOURCE}")

        source = context[self.CONTEXT_KEY_SOURCE]
        suffix = self.__DEFAULT_SUFFIX
        if self.CONTEXT_KEY_SUFFIX in context:
            suffix = context[self.CONTEXT_KEY_SUFFIX]

        files: List[str] = [os.path.join(source, file) for file in os.listdir(source) if file.endswith(suffix)]
        result: List[str] = []

        if self.CONTEXT_KEY_TARGET in context:
            target = context[self.CONTEXT_KEY_TARGET]
            for file in files:
                basename = os.path.basename(file)
                new_file = os.path.join(target, basename)
                shutil.move(file, new_file)
                result.append(new_file)
        else:
            result = files

        logger.info("[get|out] => {result}")
        return result



import logging
import shutil
from typing import Any, Dict, Optional
from tgedr.dataops.sink.sink import Sink, SinkException


logger = logging.getLogger(__name__)


class LocalFsFileSink(Sink):
    CONTEXT_SOURCE_PATH = "source"
    CONTEXT_TARGET_PATH = "target"

    def put(self, context: Optional[Dict[str, Any]] = None) -> Any:
        logger.info(f"[put|in] ({context})")

        if self.CONTEXT_SOURCE_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_SOURCE_PATH}")
        if self.CONTEXT_TARGET_PATH not in context:
            raise SinkException(f"you must provide context for {self.CONTEXT_TARGET_PATH}")

        source = context[self.CONTEXT_SOURCE_PATH]
        target = context[self.CONTEXT_TARGET_PATH]

        shutil.copy(source, target)
        logger.info("[put|out]")



import logging
from typing import Optional
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import s3fs
from tgedr.dataops.store.store import StoreException
from tgedr.dataops.store.fs_single_partition_parquet import FsSinglePartitionParquetStore


logger = logging.getLogger(__name__)


class S3FsSinglePartitionParquetStore(FsSinglePartitionParquetStore):
    @property
    def fs(self):
        if self._fs is None:
            self._fs = s3fs.S3FileSystem()
        return self._fs

    def _rmdir(self, key):
        # if self.fs.get_file_info(key).type.name == 'Directory':
        if self.fs.isdir(key):
            self.fs.delete(key, recursive=True)

    def _exists(self, key) -> bool:
        return self.fs.get_file_info(key).type.name != "NotFound"

    def get(self, key: str, use_legacy_dataset: bool = False) -> pd.DataFrame:
        logger.info(f"[get|in] ({key}, {use_legacy_dataset})")
        result = pq.read_table(key, filesystem=s3fs.S3FileSystem()).to_pandas()
        logger.info(f"[get|out] => {result}")
        return result

    def save(
        self,
        df: pd.DataFrame,
        key: str,
        partition_field: Optional[str] = None,
        append: bool = False,
        replace_partitions: bool = False,
    ):
        logger.info(f"[save|in] ({df}, {key}, {partition_field}, {append}, {replace_partitions})")
        if replace_partitions and append:
            raise StoreException(f"cannot request for replace_partitions and append at the same time")

        if append:
            pq.write_to_dataset(
                pa.Table.from_pandas(df), root_path=key, partition_cols=[partition_field], filesystem=self.fs
            )
        elif replace_partitions:
            partitions = df[partition_field].unique().tolist()
            self._remove_partitions(key, partition_field, partitions)
            pq.write_to_dataset(
                pa.Table.from_pandas(df),
                root_path=key,
                partition_cols=[partition_field],
                existing_data_behavior="delete_matching",
                filesystem=self.fs,
            )
        else:
            self.delete(key)
            pq.write_to_dataset(
                pa.Table.from_pandas(df),
                root_path=key,
                partition_cols=[partition_field],
                existing_data_behavior="delete_matching",
                filesystem=self.fs,
            )
        logger.info("[save|out]")



import abc
from typing import Any, Dict, Optional

from tgedr.dataops.processor import Processor


class ChainException(Exception):
    pass


class ChainInterface(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, "next")
            and callable(subclass.next)
            and hasattr(subclass, "execute")
            and callable(subclass.execute)
        ) or NotImplemented


class ChainMixin(abc.ABC):
    def next(self, handler: "ChainMixin") -> "ChainMixin":
        if "_next" not in self.__dict__ or self._next is None:
            self._next: "ChainMixin" = handler
        else:
            self._next.next(handler)
        return self

    @abc.abstractmethod
    def execute(self, context: Optional[Dict[str, Any]] = None) -> Any:
        raise NotImplementedError()


class ProcessorChainMixin(ChainMixin):
    def execute(self, context: Optional[Dict[str, Any]] = None) -> Any:
        self.process(context=context)
        if "_next" in self.__dict__ and self._next is not None:
            self._next.execute(context=context)


@ChainInterface.register
class ProcessorChain(ProcessorChainMixin, Processor):
    pass


@ChainInterface.register
class Chain(ChainMixin, abc.ABC):
    @abc.abstractmethod
    def execute(self, context: Optional[Dict[str, Any]] = None) -> Any:
        raise NotImplementedError()



import logging
from pyarrow import fs

from tgedr.dataops.store.fs_single_partition_parquet import FsSinglePartitionParquetStore


logger = logging.getLogger(__name__)


class LocalFsSinglePartitionParquetStore(FsSinglePartitionParquetStore):
    @property
    def fs(self):
        if self._fs is None:
            self._fs = fs.LocalFileSystem()
        return self._fs

    def _rmdir(self, key):
        if self.fs.get_file_info(key).type.name == "Directory":
            self.fs.delete_dir(key)

    def _exists(self, key) -> bool:
        return self.fs.get_file_info(key).type.name != "NotFound"



import importlib
import inspect
import logging
import os
import sys
from importlib import import_module
from typing import Dict, Any, List


logger = logging.getLogger(__name__)


class UtilsReflectionException(Exception):
    pass


class UtilsReflection:
    __MODULE_EXTENSIONS = (".py", ".pyc", ".pyo")

    @staticmethod
    def load_class(clazz: str, parent_check: type = None) -> Any:
        logger.info(f"[load_class|in] (clazz={clazz}, parent_check={parent_check})")
        type_elements = clazz.split(".")
        module = ".".join(type_elements[:-1])
        _clazz = type_elements[-1]

        result = getattr(import_module(module), _clazz)

        if not callable(result):
            raise TypeError(f"Object {_clazz} in {module} is not callable.")

        if parent_check and (not issubclass(result, parent_check)):
            raise TypeError(f"Wrong class type, it is not a subclass of {parent_check.__name__}")

        logger.info(f"[load_class|out] => {result}")
        return result

    @staticmethod
    def load_subclass_from_module(module: str, clazz: str, super_clazz: type) -> Any:
        logger.info(f"[load_subclass_from_module|in] (module={module}, clazz={clazz}, super_clazz={super_clazz})")
        result = getattr(import_module(module), clazz)

        if not callable(result):
            raise TypeError(f"Object {clazz} in {module} is not callable.")

        if super_clazz and (not issubclass(result, super_clazz)):
            raise TypeError(f"Wrong class type, it is not a subclass of {super_clazz.__name__}")

        logger.info(f"[load_subclass_from_module|out] => {result}")
        return result

    @staticmethod
    def get_type(module: str, _type: str) -> type:
        logger.info(f"[get_type|in] (module={module}, _type={_type})")
        result = None

        result = getattr(import_module(module), _type)

        logger.info(f"[get_type|out] => {result}")
        return result

    @staticmethod
    def is_subclass_of(sub_class: type, super_class: type) -> bool:
        logger.info(f"[is_subclass_of|in] ({sub_class}, {super_class})")
        result = False

        if callable(sub_class) and issubclass(sub_class, super_class):
            result = True

        logger.info(f"[is_subclass_of|out] => {result}")
        return result

    @staticmethod
    def find_module_classes(module: str) -> List[Any]:
        logger.info(f"[find_module_classes|in] ({module})")
        result = []
        for name, obj in inspect.getmembers(sys.modules[module]):
            if inspect.isclass(obj):
                result.append(obj)
        logger.info(f"[find_module_classes|out] => {result}")
        return result

    @staticmethod
    def find_class_implementations_in_package(package_name: str, super_class: type) -> Dict[str, type]:
        logger.info(f"[find_class_implementations_in_package|in] ({package_name}, {super_class})")
        result = {}

        the_package = importlib.import_module(package_name)
        pkg_path = the_package.__path__[0]
        modules = [
            package_name + "." + module.split(".")[0]
            for module in os.listdir(pkg_path)
            if module.endswith(UtilsReflection.__MODULE_EXTENSIONS) and module != "__init__.py"
        ]

        logger.info(f"[find_class_implementations_in_package] found modules: {modules}")

        for _module in modules:
            if _module not in sys.modules:
                importlib.import_module(_module)

            for _class in UtilsReflection.find_module_classes(_module):
                if UtilsReflection.is_subclass_of(_class, super_class) and _class != super_class:
                    result[_module] = _class

        logger.info(f"[find_class_implementations_in_package|out] => {result}")
        return result

    @staticmethod
    def find_package_path(package_name: str) -> str:
        logger.info(f"[find_package_path|in] ({package_name})")
        the_package = importlib.import_module(package_name)
        result = the_package.__path__[0]
        logger.info(f"[find_package_path|out] => {result}")
        return result

    @staticmethod
    def find_class_implementations(packages: str, clazz: Any) -> Dict[str, Any]:
        """
        throws UtilsReflectionException
        """
        logger.info(f"[find_class_implementations|in] ({packages}, {clazz})")
        result = {}
        _packages = [a.strip() for a in packages.split(",")]

        # find classes that extend clazz
        for pack_name in _packages:
            module_class_map = UtilsReflection.find_class_implementations_in_package(pack_name, clazz)
            for mod, _clazz in module_class_map.items():
                impl = mod.split(".")[-1]
                result[impl] = _clazz

        logger.info(f"[find_class_implementations|out] => {result}")
        return result



import os
import boto3


class S3Connector:
    def __init__(self):
        self.__resource = None
        self.__session = None
        self.__client = None

    @property
    def _session(self):
        if self.__session is None:
            if "1" == os.getenv("S3_CONNECTOR_USE_CREDENTIALS", default="0"):
                self.__session = boto3.Session(
                    aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
                    aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"],
                    region_name=os.environ["AWS_REGION"],
                )
            else:
                self.__session = boto3.Session()

        return self.__session

    @property
    def _resource(self):
        if self.__resource is None:
            self.__resource = self._session.resource("s3")
        return self.__resource

    @property
    def _client(self):
        if self.__client is None:
            self.__client = self._session.client("s3")
        return self.__client



from abc import ABC, abstractmethod
import inspect
import logging
from typing import Any, Dict, Optional


logger = logging.getLogger(__name__)


class EtlException(Exception):
    pass


"""
ETL is an abstract base class that should be extended when you want to run an ETL-like task/job
A subclass of ETL has extract, transform and load methods.

The ETL class has static utility methods that serve as an outline for the class. Use example below:

```python
class MyEtl(Etl):
    @Etl.inject_configuration
    def extract(self, MY_PARAM) -> None:
        # "MY_PARAM" should be supplied in 'configuration' dict otherwise an exception will be raised

    @Etl.inject_configuration
    def load(self, NOT_IN_CONFIG=123) -> None:
        # If you try to inject a configuration key that is NOT on the configuration dictionary
        # supplied to the constructor, it will not throw an error as long as you set a default
        # value in the method you wish to decorate
        assert NOT_IN_CONFIG == 123, "This will be ok"

```
"""


class Etl(ABC):
    def __init__(self, configuration: Optional[Dict[str, Any]] = None) -> None:
        """Initialize a new instance of ETL.

        Parameters
        ----------
        configuration : Dict[str, Any]
            source for configuration injection
        """
        self._configuration = configuration

    @abstractmethod
    def extract(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def transform(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def load(self) -> None:
        raise NotImplementedError()

    def validate_extract(self):
        """
        Optional extra checks for extract step.
        """

    def validate_transform(self):
        """
        Optional extra checks for transform step.
        """

    def run(self) -> Any:
        logger.info("[run|in]")

        self.extract()
        self.validate_extract()

        self.transform()
        self.validate_transform()

        self.load()

        logger.info("[run|out]")

    @staticmethod
    def inject_configuration(f):
        def decorator(self):
            signature = inspect.signature(f)

            missing_params = []
            params = {}
            for param in [parameter for parameter in signature.parameters if parameter != "self"]:
                if signature.parameters[param].default != inspect._empty:
                    params[param] = signature.parameters[param].default
                else:
                    params[param] = None
                    if self._configuration is None or param not in self._configuration:
                        missing_params.append(param)

                if self._configuration is not None and param in self._configuration:
                    params[param] = self._configuration[param]

            if 0 < len(missing_params):
                raise EtlException(
                    f"{type(self).__name__}.{f.__name__}: missing required configuration parameters: {missing_params}"
                )

            return f(
                self,
                *[params[argument] for argument in params],
            )

        return decorator



import logging
import os
from typing import Any, Dict, List, Optional
from tgedr.dataops.commons.s3_connector import S3Connector
from tgedr.dataops.source.source import Source, SourceException


logger = logging.getLogger(__name__)


class S3FileSource(Source, S3Connector):
    CONTEXT_KEY_SOURCE = "source"
    CONTEXT_KEY_TARGET = "target"

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        Source.__init__(self, config=config)
        S3Connector.__init__(self)

    def __derive_local_file(self, target: str, isdir: bool, file: str):
        logger.info(f"[__derive_local_file|in] ({target}, {isdir}, {file})")

        if isdir:
            result = os.path.join(target, file)
            basedir = os.path.dirname(result)
            if not os.path.exists(basedir):
                logger.info(f"[__derive_local_file] creating folder: {basedir}")
                os.mkdir(basedir)
        else:
            result = target

        logger.info(f"[__derive_local_file|out] => {result}")
        return result

    def get(self, context: Optional[Dict[str, Any]] = None) -> Any:
        logger.info(f"[get|in] ({context})")

        result: List[str] = []
        if self.CONTEXT_KEY_SOURCE not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_SOURCE}")
        if self.CONTEXT_KEY_TARGET not in context:
            raise SourceException(f"you must provide context for {self.CONTEXT_KEY_TARGET}")

        path = context[self.CONTEXT_KEY_SOURCE]
        path_elements = path.split("/")
        bucket = path_elements[0]
        key = "/".join(path_elements[1:])
        target = context[self.CONTEXT_KEY_TARGET]
        target_is_dir = os.path.isdir(target)

        objs = self._client.list_objects_v2(Bucket=bucket, Prefix=key)
        files = [entry["Key"] for entry in objs["Contents"] if not (entry["Key"]).endswith("/")]

        for file in files:
            logger.info(f"[get] found file: {file}")
            local_file = self.__derive_local_file(target=target, isdir=target_is_dir, file=file)
            logger.info(f"[get] bucket: {bucket}   key: {key}   file: {file}   local_file: {local_file}")
            self._client.download_file(Bucket=bucket, Key=file, Filename=local_file)
            result.append(local_file)

        logger.info(f"[get|out] => {result}")
        return result



import logging
import os

from setuptools import setup, find_namespace_packages

logger = logging.getLogger(__name__)
VERSION = "0.0.18"
logging.info(f"building version: {VERSION}")

setup(
    name='tgedr-dataops',
    version=VERSION,
    description='data operations related code',
    url='https://github.com/jtviegas-sandbox/dataops',
    author='joao tiago viegas',
    author_email='jtviegas@gmail.com',
    license='Unlicense',
    classifiers=[
    'Development Status :: 3 - Alpha',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10'
    ],
    keywords='data engineering mlops ml',
    include_package_data=True,
    package_dir={"": "src"},
    packages=find_namespace_packages(where="src"),
    install_requires=[
        "pyarrow==15.*",
        "pyspark==3.5",
        "s3fs==2024.2.0"
    ],
    python_requires='>=3.9',
    # package_data={'sample': ['package_data.dat'],},
)



import abc
from typing import Any, Dict, Optional

from tgedr.dataops.chain import Chain


class SinkException(Exception):
    pass


class SinkInterface(metaclass=abc.ABCMeta):
    """
    def put(self, context: Optional[Dict[str, Any]] = None) -> Any:
        raise NotImplementedError()
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, "put") and callable(subclass.put) or NotImplemented


@SinkInterface.register
class Sink(abc.ABC):
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        self._config = config

    @abc.abstractmethod
    def put(self, context: Optional[Dict[str, Any]] = None) -> Any:
        raise NotImplementedError()


@SinkInterface.register
class SinkChain(Chain, abc.ABC):
    def execute(self, context: Optional[Dict[str, Any]] = None) -> Any:
        return self.put(context=context)
